package ru.dbotthepony.mc.prng;

import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import it.unimi.dsi.fastutil.HashCommon;
import net.minecraft.util.Mth;
import net.minecraft.util.RandomSource;
import net.minecraft.world.level.levelgen.PositionalRandomFactory;
import net.minecraft.world.level.levelgen.RandomSupport;
import org.jetbrains.annotations.NotNull;

import java.util.random.RandomGenerator;

public class GJRAND64RandomSource implements RandomGenerator, RandomSource {
	public static final Codec<GJRAND64RandomSource> CODEC = RecordCodecBuilder.create(it -> {
		return it.group(
			Codec.LONG.fieldOf("s0").forGetter(o -> o.s0),
			Codec.LONG.fieldOf("s1").forGetter(o -> o.s1),
			Codec.LONG.fieldOf("s2").forGetter(o -> o.s2),
			Codec.LONG.fieldOf("s3").forGetter(o -> o.s3)
		).apply(it, GJRAND64RandomSource::new);
	});

	private long s0;
	private long s1;
	private long s2;
	private long s3;

	private GJRAND64RandomSource(
		long s0,
		long s1,
		long s2,
		long s3
	) {
		this.s0 = s0;
		this.s1 = s1;
		this.s2 = s2;
		this.s3 = s3;
	}

	public GJRAND64RandomSource() {
		reinitialize(RandomSupport.generateUniqueSeed(), RandomSupport.generateUniqueSeed());
	}

	public GJRAND64RandomSource(long seed) {
		reinitialize(seed);
	}

	public GJRAND64RandomSource(long seed0, long seed1) {
		reinitialize(seed0, seed1);
	}

	public GJRAND64RandomSource(RandomSupport.Seed128bit seed) {
		this(seed.seedHi(), seed.seedLo());
	}

	private void reinitialize(long seed) {
		reinitialize(seed, 0L);
	}

	private void reinitialize(long seed0, long seed1) {
		s0 = seed0;
		s1 = seed1;
		s2 = 2000001L;
		s3 = 0L;

		for (int i = 0; i < 14; i++) {
			round();
		}
	}

	protected final long round() {
		s1 += s2;
		s0 = Long.rotateLeft(s0, 32);
		s2 ^= s1;
		s3 += 0x55aa96a5L;
		s0 += s1;
		s2 = Long.rotateLeft(s2, 23);
		s1 ^= s0;
		s0 += s2;
		s1 = Long.rotateLeft(s1, 19);
		s2 += s0;
		s1 += s3;
		return s0;
	}

	@Override
	public @NotNull RandomSource fork() {
		return new GJRAND64RandomSource(nextLong(), nextLong());
	}

	@Override
	public @NotNull PositionalRandomFactory forkPositional() {
		return new Positional(nextLong(), nextLong());
	}

	@Override
	public void setSeed(long l) {
		reinitialize(l);
	}

	@Override
	public int nextInt() {
		return RandomGenerator.super.nextInt();
	}

	@Override
	public int nextInt(int bound) {
		return RandomGenerator.super.nextInt(bound);
	}

	@Override
	public int nextInt(int origin, int bound) {
		return RandomGenerator.super.nextInt(origin, bound);
	}

	@Override
	public float nextFloat() {
		return RandomGenerator.super.nextFloat();
	}

	@Override
	public boolean nextBoolean() {
		return RandomGenerator.super.nextBoolean();
	}

	@Override
	public double nextDouble() {
		return RandomGenerator.super.nextDouble();
	}

	@Override
	public double nextGaussian() {
		return RandomGenerator.super.nextGaussian();
	}

	@Override
	public long nextLong() {
		return round();
	}

	public static class Positional implements PositionalRandomFactory {
		private final long seed0;
		private final long seed1;

		public Positional(long seed0, long seed1) {
			this.seed0 = seed0;
			this.seed1 = seed1;
		}

		@Override
		public @NotNull RandomSource fromHashOf(@NotNull String s) {
			long seed = HashCommon.murmurHash3((long) s.hashCode() & 0xFFFFFFFFL);
			return new GJRAND64RandomSource(seed0 ^ Long.rotateLeft(seed, 32), seed1 ^ seed);
		}

		@Override
		public @NotNull RandomSource fromSeed(long l) {
			return new GJRAND64RandomSource(l);
		}

		@Override
		public @NotNull RandomSource at(int x, int y, int z) {
			long seed = Mth.getSeed(x, y, z);
			return new GJRAND64RandomSource(seed0 ^ Long.rotateLeft(seed, 32), seed1 ^ seed);
		}

		@Override
		public void parityConfigString(@NotNull StringBuilder stringBuilder) {
			throw new UnsupportedOperationException();
		}
	}
}
