package ru.dbotthepony.mc.prng.mixin;

import net.minecraft.core.BlockPos;
import net.minecraft.core.Holder;
import net.minecraft.core.RegistryAccess;
import net.minecraft.resources.ResourceKey;
import net.minecraft.util.RandomSource;
import net.minecraft.util.profiling.ProfilerFiller;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.dimension.DimensionType;
import net.minecraft.world.level.storage.WritableLevelData;
import org.spongepowered.asm.mixin.Final;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Overwrite;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Inject;
import org.spongepowered.asm.mixin.injection.Redirect;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfo;
import ru.dbotthepony.mc.prng.BetterRandom;
import ru.dbotthepony.mc.prng.GJRAND64RandomSource;

import java.util.function.Supplier;

@Mixin(value = Level.class, priority = 10000) // ensure this mixin is applied over other "regular" mixins
public abstract class LevelMixin {
	@Shadow
	@Final
	private Thread thread;

	// does not work with C2ME, Mixin crashes
	/*
	@Redirect(
		method = "<init>",
		at = @At(
			value = "INVOKE",
			target = "Lnet/minecraft/util/RandomSource;create()Lnet/minecraft/util/RandomSource;"
		)
	)
	public RandomSource ensureBetterRandom() {
		return BetterRandom.createWorldRandom(() -> thread);
	}
	 */

	@Shadow
	public RandomSource random;

	@Inject(
		method = "<init>(Lnet/minecraft/world/level/storage/WritableLevelData;Lnet/minecraft/resources/ResourceKey;Lnet/minecraft/core/RegistryAccess;Lnet/minecraft/core/Holder;Ljava/util/function/Supplier;ZZJI)V",
		at = @At("TAIL")
	)
	public void constructorMix(
		WritableLevelData levelData,
		ResourceKey<Level> dimension,
		RegistryAccess registryAccess,
		Holder<DimensionType> dimensionTypeRegistration,
		Supplier<ProfilerFiller> profiler,
		boolean isClientSide,
		boolean isDebug,
		long biomeZoomSeed,
		int maxChainedNeighborUpdates,
		CallbackInfo info
	) {
		if (!(this.random instanceof GJRAND64RandomSource))
			this.random = BetterRandom.createWorldRandom(() -> thread);
	}


	@Overwrite
	public BlockPos getBlockRandomPos(int x, int y, int z, int yMask) {
		long value = ((Level) (Object) this).random.nextLong();
		int selectX = (int) (value & 15);
		int selectZ = (int) ((value >>> 8) & 15);
		int selectY = (int) ((value >>> 16) & yMask);
		return new BlockPos(selectX + x, selectY + y, selectZ + z);
	}
}
